#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <fenv.h>
#include <string.h>

struct func {
	char *name;
	int args;
	double (*f)();
} funcs[] = {
	{"acos",  1, acos},
	{"asin",  1, asin},
	{"atan",  1, atan},
	{"atan2", 2, atan2},
	{"ceil",  1, ceil},
	{"cos",   1, cos},
	{"cosh",  1, cosh},
	{"exp",   1, exp},
	{"expm1", 1, expm1},
	{"floor", 1, floor},
	{"fmod",  2, fmod},
	{"hypot", 2, hypot},
	{"log",   1, log},
	{"log10", 1, log10},
	{"log1p", 1, log1p},
	{"log2",  1, log2},
	{"pow",   2, pow},
	{"sin",   1, sin},
	{"sinh",  1, sinh},
	{"sqrt",  1, sqrt},
	{"tan",   1, tan},
	{"tanh",  1, tanh},
};
static int nfuncs = sizeof funcs / sizeof *funcs;

struct test {
	struct func *func;
	int round;
	int except;
	int exceptopt;
	double in1;
	double in2;
	double out;
};

static double tod(uint32_t *a) {
	union { double x; uint64_t n; } u;

	u.n = ((uint64_t)a[0] << 32) | a[1];
	return u.x;
}

int scantest(struct test *t) {
	char name[32];
	char round[32];
	char cmp[32];
	char except[32];
	char buf[512];
	uint32_t a[6];
	int i;

	if (!fgets(buf, sizeof buf, stdin))
		return 0;

	if (sscanf(buf, "%30s %30s %30s %30s %x %x %x %x %x %x",
	     name, round, cmp, except, a+0, a+1, a+2, a+3, a+4, a+5) < 8)
		return 0;

	for (i = 0; i < nfuncs; i++)
		if (strcmp(funcs[i].name, name) == 0)
			break;
	if (i == nfuncs)
		return 0;
	t->func = funcs+i;
	t->round = round[0] == 'n';

	t->except = t->exceptopt = 0;
	for (i = 0; i < sizeof except; i++) {
		int f;

		if (except[i] == '\0')
			break;
		switch (except[i]) {
		case 'v': f = FE_INVALID; break;
		case 'x': f = FE_INEXACT; break;
		case 'o': f = FE_OVERFLOW; break;
		case 'u': f = FE_UNDERFLOW; break;
		case 'd': f = FE_DIVBYZERO; break;
		default:
			continue;
		}
		if (i > 0 && except[i-1]=='?')
			t->exceptopt |= f;
		else
			t->except |= f;
	}

	t->in1 = tod(a);
	t->in2 = 0;
	if (t->func->args == 1) {
		t->out = tod(a+2);
	} else {
		t->in2 = tod(a+2);
		t->out = tod(a+4);
	}

	if (strcmp(cmp, "uo") == 0)
		t->out = 0.0/0.0;
	return 1;
}

int main(){
	struct test t;
	union {double x; uint64_t n;} got, want;
	int64_t k;
	int n=0, err=0, err2=0, err99=0;

	while (!feof(stdin)) {
		if(!scantest(&t)) {
//			puts("scan fail");
			continue;
		}
		if (!t.round)
			continue;
		if (t.func->args == 1)
			got.x = t.func->f(t.in1);
		else
			got.x = t.func->f(t.in1, t.in2);
		want.x = t.out;

		if (got.n != want.n &&
		    (!isnan(got.x) || !isnan(want.x))) {
			err++;
			// hack: mostly error in ulp
			k = (int64_t)(got.n - want.n);
			if (k >= 100)
				k = 99;
			if (k <= -100)
				k = -99;
			if (k > 1 || k < -1)
				err2++;
			if (k == 99 || k == -99)
				err99++;
			printf("%3d ulp %16llx %16llx   %s %a %a %a %a\n",
			  (int)k, want.n, got.n, t.func->name, t.in1, t.in2, t.out, got.x);
		}
		n++;
	}
	printf("all: %d fail: %d failbad: %d failepic: %d\n", n, err, err2, err99);
	return 0;
}
